
from rich.console import Console
from rich.themes import Theme


theme = Theme({
    "err": "bold bright_red",       # standard error
    "plog": "bold bright_green",    # compilation started or finished
    "cerror": "bold deep_pink2",    # compilation error
    "lilog": "bold bright_cyan",    # linking log
    "hinfo": "steel_blue1",         
    "abort": "orange_red1",
    "sigint": "magenta2",
})

console = Console(theme=theme)


import os
import time
import psutil
import tkinter as tk


from .. import console
from .make import make_app
from ..paths import PROJ_DIRS
from ..config import get_proj_config
from ..utils import check_valid_proj_and_zig_installed


# Compiles the project and Runs the executable.
def run_prog(remake: bool, show_metrics: bool = False) -> None:

    # validate project and requirements
    if not check_valid_proj_and_zig_installed():
        return

    # get project config
    proj_config = get_proj_config()

    # Create program path
    program = os.path.join(PROJ_DIRS["DEBUG_BIN_DIR"], proj_config['project']['name'])

    # Recompile if remake true
    if remake:
        if not make_app():
            return

    if os.path.exists(program):

        process = psutil.Popen([program])

        window = tk.Tk()
        window.geometry("480x360")
        window.resizable(False, False)

        stats = tk.Label(window, justify="left")
        stats.pack(anchor='w')

        if show_metrics:
            try:

                CPU_limit = proj_config["debug"]["limit"]["cpu"]
                RAM_limit = proj_config["debug"]["limit"]["ram"]
                Thread_limit = proj_config["debug"]["limit"]["thread"]

                while process.status() not in [psutil.STATUS_STOPPED, psutil.STATUS_ZOMBIE, psutil.STATUS_DEAD]:
                    
                    cpu_percent = process.cpu_percent()

                    mem_info = process.memory_info()
                    rss = mem_info.rss / (1024 * 1024) # in MB

                    threads = process.num_threads()

                    if cpu_percent > CPU_limit and CPU_limit != 0 :
                        console.print("\n[abort]CPU limit exceeded[/abort]")
                        break
                    elif rss > RAM_limit and RAM_limit != 0 :
                        console.print("\n[abort]RAM limit exceeded[/abort]")
                        break
                    elif threads > Thread_limit and Thread_limit != 0 :
                        console.print("\n[abort]Thread limit exceeded[/abort]")
                        break

                    time.sleep(0.5)

                    stats.config(text=f"""
    Process Name: {process.name()}
    Process ID: {process.pid}

    CPU Usage: {cpu_percent}%
    Memory Usage: {process.memory_percent():.2f}%
        [RSS]  {rss:.2f} MB
        [VMS]  {process.memory_info().vms / (1024 * 1024):.2f} MB
    
    Thread Count: {threads}

    Read Bytes: {process.io_counters().read_bytes / (1024 * 1024):.2f} MB
    Write Bytes: {process.io_counters().write_bytes / (1024 * 1024):.2f} MB
""")

                    window.update_idletasks()
                    window.update()

            except KeyboardInterrupt:
                console.print("\n[sigint]User Interrupt[/sigint]")
            except Exception as e:
                console.print("[err]Error occured[/err] ->", e)
            finally:
                process.kill()
                window.quit()
                console.print("[sigint]Process Terminated[/sigint]")
    else:
        console.print("[err]Error[/err] -> Executable not found")


import os


from .. import console
from ..paths import PROJ_DIRS, cwd, MODIF_LOG_FILE
from ..config import create_proj_config_file

# Creates a new project
# with the given name
# @param project_name: str
def create_cmd(name: str, its_package: bool) -> None:

    if os.listdir(cwd):
        console.print("[err]Error[/err] directory not empty!")
        return

    with open('.gitignore', 'w') as f:
            f.write('''
.candie/
build/
debug/
include/
lib/
package/
test/
''')

    create_proj(project_name=name)
    console.print("[bright_white]Project created.[/bright_white]")

    with open(MODIF_LOG_FILE, 'w') as f:
        f.write('{}')



def create_proj(project_name: str):
    create_proj_config_file(project_name)
    for dir in PROJ_DIRS.values():
        os.makedirs(dir, exist_ok=True)

    with open(os.path.join(PROJ_DIRS["SRC_DIR"], 'main.c'), 'w') as f:
        f.write('''
#include<stdio.h>

int main() {
    printf("Hello, World!");
    return 0;
}
''')


# def create_pkg(package_name: str):
#     create_pkg_config_file(package_name)
#     for dir in PKG_DIRS.values():
#         os.makedirs(dir, exist_ok=True)

#     with open(os.path.join(PKG_DIRS["PACKAGE_HEADERS_DIR"], f'{package_name}.h'), 'w') as f:
#         f.write(f'''
# #ifndef _{package_name.upper()}_H_
# #define _{package_name.upper()}_H_
                
# int add(int a, int b);

# #endif
# ''')

#     with open(os.path.join(PKG_DIRS["PACKAGE_DIR"], f'{package_name}.c'), 'w') as f:
#         f.write(f'''
# #include "HEADERS/{package_name}.h"

# int add(int a, int b) {{
#     return a + b;
# }}
# ''')

#     with open(os.path.join(PKG_DIRS["SRC_DIR"], 'main.c'), 'w') as f:
#         f.write(f'''
# #include<stdio.h>
# #include<{package_name}.h>

# int main() {{
#     printf("%d", add(5 + 5));
#     return 0;
# }}
# ''')


import os
import json
import toml
import shutil

from .. import console
from ..paths import PKG_INDEX_FILE, PROJ_CONFIG_FILE

# Removes all package files from the project directory
# @param package_name (str): name of the package
def remove_pkg(package_name: str) -> None:

    pkg_index_data: list[dict] = []

    if os.path.exists(PKG_INDEX_FILE):
        with open(PKG_INDEX_FILE, 'r') as f:
            pkg_index_data = json.load(f)

    pkg_to_remove: dict = None

    for pkg in pkg_index_data:
        for pkg_name in pkg.keys():
            if pkg_name == package_name:
                pkg_to_remove = pkg
            break

    if not pkg_to_remove:
        console.print(f"[dim]Error[/dim] -> Package {package_name} is not added to this project.")
        return
        
    for pkg_dirs in pkg.values():
        for dir in pkg_dirs:
            for path in dir:
                if os.path.isfile(path):
                    os.remove(path)
                elif os.path.isdir(path):
                    shutil.rmtree(path)

    # pkg_index_data.remove(pkg)
    remove_pkg_from_requirements(package_name)

    with open(PKG_INDEX_FILE, 'w') as f:
        json.dump(pkg_index_data, f, indent=4)

    console.print(f"[dim]{package_name}[/dim] [hinfo]Package removed[/hinfo]")




# Removes the package from the requirements table in the project config
# @param: (str) Package name
def remove_pkg_from_requirements(package_name: str) -> None:

    with open(PROJ_CONFIG_FILE, "r") as f:
        data = toml.load(f)

    for pkg in data.get("requirements", {}).get("package", []):
        if pkg["name"] == package_name:
            data.setdefault("requirements", {}).setdefault("package", []).remove(pkg)

    with open(PROJ_CONFIG_FILE, "w") as f:
        toml.dump(data, f)


import os


from ..paths import PROJ_DIRS
from ..config import Package, get_proj_config, read_pc_file
from ..utils import zig_link, get_src_files, check_valid_proj_and_zig_installed


# Builds the project
def build_proj():

    if not check_valid_proj_and_zig_installed():
        return
    
    proj_config = get_proj_config()

    cflags: list[str] = []
    libs: list[str] = []

    for pc_file in os.scandir(PROJ_DIRS["DEBUG_LIB_PKGCONFIG_DIR"]):
        pkg: Package = read_pc_file(pc_file.path)
        if pkg:
            cflags.append(pkg.cflags.replace('\"', '').replace('${includedir}', PROJ_DIRS["INCLUDE_DIR"]))
            libs.append(pkg.libs.replace('\"', '').replace('${libdir}', PROJ_DIRS["DEBUG_LIB_DIR"]))

    
    for target in proj_config['build']['target']:
        zig_link(
            input_files=get_src_files(), 
            output_path=os.path.join(PROJ_DIRS["BUILD_DIR"], f'{proj_config['project']['name']}-{target['arch']}-{target['os']}'),
            cflags=cflags,
            libs=libs,
            target=f'{target['arch']}-{target['os']}',
            additional_flags=proj_config['build']['flags']
        )


import os
import json

from threading import Thread


from .. import console
from ..paths import PROJ_DIRS, MODIF_LOG_FILE
from ..config import get_proj_config, read_pc_file, Package
from ..utils import zig_compile, zig_link, get_object_files, get_src_files, check_valid_proj_and_zig_installed


# Makes an executable file
# compiles all source files and cache them
# links all the object files in to an executable
def make_app(verbose: bool = False) -> bool:

    # validate project and requirements
    if not check_valid_proj_and_zig_installed():
        return

    # Get project config
    proj_config = get_proj_config()

    logs = get_logs()    # Logs

    # Object files and src files
    o_files: dict = get_object_files()       # object files
    src_files: list[str] = get_src_files()   # source files

    if len(src_files) <= 0: 
        console.print("[err]Error[/err] -> Empty src directory.")
        return False

    # Options to provide while compiling and linking
    cflags: list[str] = []   # Compilation flags
    libs: list[str] = []     # Linker flags


    # Collect cflags and libs from all package config files
    for pc_file in os.scandir(PROJ_DIRS["DEBUG_LIB_PKGCONFIG_DIR"]):
        pkg: Package = read_pc_file(pc_file.path)
        if pkg:
            cflags.append(pkg.cflags.replace('\"', '').replace('${includedir}', PROJ_DIRS["INCLUDE_DIR"]))
            libs.append(pkg.libs.replace('\"', '').replace('${libdir}', PROJ_DIRS["DEBUG_LIB_DIR"]))


    # list containing threads of compile_src functions
    compile_queue: list[Thread] = []

    # Checks if error occurs while compilation
    compile_status: list[bool] = []

    # For all source files check if the file is newly created or modified.
    # newly created files will be compiled
    # and modified files will be re-compiled
    for src_path in src_files:

        # Get the last modified time stat
        file_last_modified = os.stat(src_path).st_mtime

        # hex encoded filename
        cache_file = src_path.encode('utf-8').hex() + '.o'
        
        # if cached file is not present in cache directory
        # or if it is modified
        if cache_file not in o_files or src_path in logs and logs[src_path] != file_last_modified:
            compile_queue.append(
                Thread(
                    target=compile_src, args=({
                        'logs': logs,
                        'file_last_modified': file_last_modified
                    }, src_path, cflags, compile_status, proj_config['debug']['Cflags'])
                )
            )

    # start compilation processes
    for process in compile_queue:
        process.start()

    # Checks if the source file for the compiled object exists.
    # if not then deletes the cached object file
    for o_file, o_file_path in o_files.items():
        file = bytes.fromhex(o_file.replace('.o', '')).decode('utf-8')
        if file not in src_files:
            os.remove(o_file_path)
    
    # Checks if file in log data exists if not then deletes that log
    for file in list(logs.keys()):
        if not os.path.exists(file):
            del logs[file]

    # wait for compilation to complete
    for process in compile_queue:
        process.join()

    # Update the log file
    update_logs(logs)

    # Link all the object files and create executable
    if False not in compile_status:
        zig_link(
            input_files=[*get_object_files().values()], 
            output_path=os.path.join(PROJ_DIRS["DEBUG_BIN_DIR"], proj_config['project']['name']),
            libs=libs,
            additional_flags=proj_config['debug']['Lflags']
        )
        return True
    return False

# Reads the log file and returns its contents
# @returns: dict of logs
def get_logs() -> dict:
    logs: dict = {}
    try:
        with open(MODIF_LOG_FILE, 'r') as f:
            logs = json.load(f)
    except FileNotFoundError:
        console.print(f"[err]Error[/err] [dim](while reading logs)[/dim] -> File '{MODIF_LOG_FILE}' not found.")
    except json.JSONDecodeError as e:
        console.print(f"[err]Error[/err] [dim](while reading logs)[/dim] -> Invalid JSON syntax in file '{MODIF_LOG_FILE}': {e}")
    except Exception as e:
        console.print(f"[err]Error[/err] -> {e}")
    return logs


# Ovveride the log file with new log_content
# @param log_content (dict): new log content
def update_logs(log_content: dict) -> None:
    with open(MODIF_LOG_FILE, 'w') as f:
        json.dump(log_content, f, indent=2)


# Compilation process function
# this will execute in thread
def compile_src(log_data: dict, src_path: str, cflags: list, compile_result: list[bool], additonal_flags: str = "") -> None: 
    if not zig_compile(src_path, cflags, additional_flags=additonal_flags): # compile the file
        compile_result.append(False)
        return
    
    log_data['logs'][src_path] = log_data['file_last_modified']
    compile_result.append(True)


from setuptools import setup, find_packages

setup(
    name='candie.kit',
    version='1.1.9',
    packages=find_packages(),
    py_modules=[],
    include_package_data=True,
    install_requires=['typer', 'toml', 'psutil'],
    entry_points={
        'console_scripts': [
            'candie=candie.cli:candie_exec',
        ],
    },
    author='Yash Tajne',
    # author_email='',
    description='A CLI Build tool for C/C++ projects',
    url='https://github.com/yashtajne/candie',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    license='MIT License',
    python_requires='>=3.6',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown'
)